from ...flow import Flowable, flow

from .resource import Resource


class Api(Flowable):
    """
    Class for a Kubernetes API.
    """
    def __init__(self, client, api_version):
        self._client = client
        self._api_version = api_version
        self._resources = None

    def get_flow_executor(self):
        """
        Returns the flow executor to use.
        """
        return self._client.get_flow_executor()

    @property
    def api_version(self):
        """
        Returns the API version for the API.
        """
        return self._api_version

    def _ensure_resources(self):
        """
        Ensures that the resources have been loaded.
        """
        if self._resources is None:
            prefix = "/apis" if "/" in self._api_version else "/api"
            response = yield self._client.get(f"{prefix}/{self._api_version}")
            self._resources = { r["name"]: r for r in response.json()["resources"] }
        return self._resources

    @flow
    def resource(self, name):
        """
        Returns a resource for the given name.

        The given name can be either the plural name, the singular name or the kind.
        Lookups by plural name will be faster as that is the key that is indexed.
        """
        resources = yield self._ensure_resources()
        # First try a lookup by plural name
        try:
            resource = resources[name]
        except KeyError:
            # Then try a lookup by singular name or kind
            try:
                resource = next(
                    r
                    for r in resources.values()
                    if r["kind"] == name or r["singularName"] == name
                )
            except StopIteration:
                raise ValueError(f"API '{self._api_version}' has no resource '{name}'")
        return Resource(
            self._client,
            self._api_version,
            resource["name"],
            resource["kind"],
            resource["namespaced"]
        )



from .client import ResourceSpec


Binding = ResourceSpec("v1", "bindings", "Binding", True)
ComponentStatus = ResourceSpec("v1", "componentstatuses", "ComponentStatus", False)
ConfigMap = ResourceSpec("v1", "configmaps", "ConfigMap", True)
Endpoints = ResourceSpec("v1", "endpoints", "Endpoints", True)
Event = ResourceSpec("v1", "events", "Event", True)
LimitRange = ResourceSpec("v1", "limitranges", "LimitRange", True)
Namespace = ResourceSpec("v1", "namespaces", "Namespace", False)
Node = ResourceSpec("v1", "nodes", "Node", False)
PersistentVolumeClaim = ResourceSpec("v1", "persistentvolumeclaims", "PersistentVolumeClaim", True)
PersistentVolume = ResourceSpec("v1", "persistentvolumes", "PersistentVolume", False)
Pod = ResourceSpec("v1", "pods", "Pod", True)
PodTemplate = ResourceSpec("v1", "podtemplates", "PodTemplate", True)
ReplicationController = ResourceSpec("v1", "replicationcontrollers", "ReplicationController", True)
ResourceQuota = ResourceSpec("v1", "resourcequotas", "ResourceQuota", True)
Secret = ResourceSpec("v1", "secrets", "Secret", True)
ServiceAccount = ResourceSpec("v1", "serviceaccounts", "ServiceAccount", True)
Service = ResourceSpec("v1", "services", "Service", True)

MutatingWebhookConfiguration = ResourceSpec(
    "admissionregistration.k8s.io/v1",
    "mutatingwebhookconfigurations",
    "MutatingWebhookConfiguration",
    False
)
ValidatingWebhookConfiguration = ResourceSpec(
    "admissionregistration.k8s.io/v1",
    "validatingwebhookconfigurations",
    "ValidatingWebhookConfiguration",
    False
)

CustomResourceDefinition = ResourceSpec(
    "apiextensions.k8s.io/v1",
    "customresourcedefinitions",
    "CustomResourceDefinition",
    False
)

APIService = ResourceSpec("apiregistration.k8s.io/v1", "apiservices", "APIService", False)

ControllerRevision = ResourceSpec("apps/v1", "controllerrevisions", "ControllerRevision", True)
DaemonSet = ResourceSpec("apps/v1", "daemonsets", "DaemonSet", True)
Deployment = ResourceSpec("apps/v1", "deployments", "Deployment", True)
ReplicaSet = ResourceSpec("apps/v1", "replicasets", "ReplicaSet", True)
StatefulSet = ResourceSpec("apps/v1", "statefulsets", "StatefulSet", True)

TokenReview = ResourceSpec("authentication.k8s.io/v1", "tokenreviews", "TokenReview", False)

LocalSubjectAccessReview = ResourceSpec(
    "authorization.k8s.io/v1",
    "localsubjectaccessreviews",
    "LocalSubjectAccessReview",
    True
)
SelfSubjectAccessReview = ResourceSpec(
    "authorization.k8s.io/v1",
    "selfsubjectaccessreviews",
    "SelfSubjectAccessReview",
    False
)
SelfSubjectRulesReview = ResourceSpec(
    "authorization.k8s.io/v1",
    "selfsubjectrulesreviews",
    "SelfSubjectRulesReview",
    False
)
SubjectAccessReview = ResourceSpec(
    "authorization.k8s.io/v1",
    "subjectaccessreviews",
    "SubjectAccessReview",
    False
)

HorizontalPodAutoscaler = ResourceSpec(
    "autoscaling/v1",
    "horizontalpodautoscalers",
    "HorizontalPodAutoscaler",
    True
)

CronJob = ResourceSpec("batch/v1", "cronjobs", "CronJob", True)
Job = ResourceSpec("batch/v1", "jobs", "Job", True)

CertificateSigningRequest = ResourceSpec(
    "certificates.k8s.io/v1",
    "certificatesigningrequests",
    "CertificateSigningRequest",
    False
)

Lease = ResourceSpec("coordination.k8s.io/v1", "leases", "Lease", True)

EndpointSlice = ResourceSpec("discovery.k8s.io/v1", "endpointslices", "EndpointSlice", True)

FlowSchema = ResourceSpec(
    "flowcontrol.apiserver.k8s.io/v1beta1",
    "flowschemas",
    "FlowSchema",
    False
)
PriorityLevelConfiguration = ResourceSpec(
    "flowcontrol.apiserver.k8s.io/v1beta1",
    "prioritylevelconfigurations",
    "PriorityLevelConfiguration",
    False
)

IngressClass = ResourceSpec("networking.k8s.io/v1", "ingressclasses", "IngressClass", False)
Ingress = ResourceSpec("networking.k8s.io/v1", "ingresses", "Ingress", True)
NetworkPolicy = ResourceSpec("networking.k8s.io/v1", "networkpolicies", "NetworkPolicy", True)

RuntimeClass = ResourceSpec("node.k8s.io/v1", "runtimeclasses", "RuntimeClass", False)

PodDisruptionBudget = ResourceSpec(
    "policy/v1",
    "poddisruptionbudgets",
    "PodDisruptionBudget",
    True
)
PodSecurityPolicy = ResourceSpec(
    "policy/v1beta1",
    "podsecuritypolicies",
    "PodSecurityPolicy",
    False
)

ClusterRoleBinding = ResourceSpec(
    "rbac.authorization.k8s.io/v1",
    "clusterrolebindings",
    "ClusterRoleBinding",
    False
)
ClusterRole = ResourceSpec("rbac.authorization.k8s.io/v1", "clusterroles", "ClusterRole", False)
RoleBinding = ResourceSpec("rbac.authorization.k8s.io/v1", "rolebindings", "RoleBinding", True)
Role = ResourceSpec("rbac.authorization.k8s.io/v1", "roles", "Role", True)

PriorityClass = ResourceSpec("scheduling.k8s.io/v1", "priorityclasses", "PriorityClass", False)

CSIDriver = ResourceSpec("storage.k8s.io/v1", "csidrivers", "CSIDriver", False)
CSINode = ResourceSpec("storage.k8s.io/v1", "csinodes", "CSINode", False)
CSIStorageCapacity = ResourceSpec(
    "storage.k8s.io/v1beta1",
    "csistoragecapacities",
    "CSIStorageCapacity",
    True
)
StorageClass = ResourceSpec("storage.k8s.io/v1", "storageclasses", "StorageClass", False)
VolumeAttachment = ResourceSpec(
    "storage.k8s.io/v1",
    "volumeattachments",
    "VolumeAttachment",
    False
)



import json

import httpx

from ... import rest


class ListResponseIterator(rest.ListResponseIterator):
    """
    Iterator for list responses.
    """
    def __init__(self, client, resource, params):
        super().__init__(client, resource, params)
        self._resource_version = None

    @property
    def resource_version(self):
        """
        The last seen resource version.
        """
        return self._resource_version

    def _extract_list(self, response):
        # Store the resource version from the response as well as returning the data
        data = response.json()
        self._resource_version = data["metadata"]["resourceVersion"]
        return data["items"]


class WatchEvents(rest.TextStreamIterator):
    """
    Stream iterator that yields watch events.
    """
    def __init__(self, client, path, params, initial_resource_version):
        params = params.copy()
        params.update({
            "watch": 1,
            "resourceVersion": initial_resource_version,
            "allowWatchBookmarks": "true",
        })
        super().__init__(client, "GET", path, params = params, timeout = None)

    def _process_chunk(self, chunk):
        event = json.loads(chunk)
        # Each event contains a resourceVersion, which we track so that we can restart
        # the watch from that version if it fails
        resource_version = event["object"]["metadata"]["resourceVersion"]
        self._request_kwargs["params"]["resourceVersion"] = resource_version
        # Bookmark events are just for us to save a resource version
        # They should not be emitted
        if event["type"] == "BOOKMARK":
            raise self.SuppressItem
        else:
            return event

    def _should_resume(self, exception):
        # In the case where an event is not valid JSON or we get a RemoteProtocolError,
        # it is likely that the API server is still healthy but the read has just timed out
        #
        # In this case, we should be able to restart the watch from the last known version
        #
        # In all other cases, the exception should be allowed to bubble, e.g.:
        #   * The API server responds with 410 Gone, in which case the watch needs
        #     to restart from scratch
        #   * We cannot connect to the API server
        return (
            not exception or
            isinstance(exception, json.JSONDecodeError) or
            isinstance(exception, httpx.RemoteProtocolError)
        )



import copy
import enum

from ... import rest
from ...flow import flow

from .errors import ApiError
from .iterators import ListResponseIterator, WatchEvents


#: Sentinel object indicating that the presence of a label is required with any value
PRESENT = object()
#: Sentinel object indicating that a label must not be present
ABSENT = object()


class DeletePropagationPolicy(str, enum.Enum):
    """
    Enumeration of possible delete policies.
    """
    BACKGROUND = "Background"
    FOREGROUND = "Foreground"
    ORPHAN     = "Orphan"


class Resource(rest.Resource):
    """
    Class for Kubernetes REST resources.
    """
    __iterator_class__ = ListResponseIterator

    def __init__(self, client, api_version, name, kind, namespaced):
        super().__init__(client, name)
        self._api_version = api_version
        self._kind = kind
        self._namespaced = namespaced

    def _prepare_path(self, id = None, params = None):
        params = params or {}
        namespace = params.pop("namespace", None) or self._client.default_namespace
        all_namespaces = params.pop("all_namespaces", False)
        if "labels" in params:
            # Add the selectors for the labels to any explicit label selector
            # This allows the use of set selectors with the labels keyword
            if "labelSelector" in params:
                label_selectors = [params["labelSelector"]]
            else:
                label_selectors = []
            for k, v in params.pop("labels").items():
                if v is PRESENT:
                    label_selectors.append(k)
                elif v is ABSENT:
                    label_selectors.append(f"!{k}")
                elif isinstance(v, (list, tuple)):
                    values_text = v.join(",")
                    label_selectors.append(f"{k} in ({values_text})")
                else:
                    label_selectors.append(f"{k}={v}")
            params["labelSelector"] = ",".join(label_selectors)
        if "fields" in params:
            if "fieldSelector" in params:
                field_selectors = [params["fieldSelector"]]
            else:
                field_selectors = []
            field_selectors.extend(
                f"{k}={v}"
                for k, v in params.pop("fields").items()
            )
            params["fieldSelector"] = ",".join(field_selectors)
        # Begin with either /api or /apis depending whether the api version is the core API
        prefix = "/apis" if "/" in self._api_version else "/api"
        if self._namespaced and not all_namespaces:
            path_namespace = f"/namespaces/{namespace}"
        else:
            path_namespace = ""
        path, _ = super()._prepare_path(id)
        return f"{prefix}/{self._api_version}{path_namespace}{path}", params

    def _prepare_data(self, data, id = None, params = None):
        data = copy.deepcopy(data)
        # Update the data with the known api version and kind
        data.setdefault("apiVersion", self._api_version)
        data.setdefault("kind", self._kind)
        # Set the name in metadata to the given id
        if id:
            data.setdefault("metadata", {}).update(name = id)
        return data

    def create(self, data, /, namespace = None):
        namespace = namespace or data.get("metadata", {}).get("namespace")
        return super().create(data, namespace = namespace)

    def fetch(self, id, /, namespace = None, **params):
        return super().fetch(id, namespace = namespace, **params)

    def replace(self, id, data, /, namespace = None):
        namespace = namespace or data.get("metadata", {}).get("namespace")
        return super().replace(id, data, namespace = namespace)

    def patch(self, id, data, /, namespace = None):
        namespace = namespace or data.get("metadata", {}).get("namespace")
        return super().patch(id, data, namespace = namespace)

    @flow
    def json_patch(self, id, data, /, namespace = None):
        """
        Patches the specified instance with the given data, treated as a JSON Patch.
        """
        yield self._ensure_initialised()
        path, params = self._prepare_path(id, { "namespace": namespace })
        response = yield self._client.patch(
            path,
            json = data,
            params = params,
            headers = { "Content-Type": "application/json-patch+json" }
        )
        return self._wrap_instance(self._extract_one(response))

    @flow
    def json_merge_patch(self, id, data, /, namespace = None):
        """
        Patches the specified instance with the given data, treated as a JSON Patch.
        """
        yield self._ensure_initialised()
        path, params = self._prepare_path(id, { "namespace": namespace })
        response = yield self._client.patch(
            path,
            json = data,
            params = params,
            headers = { "Content-Type": "application/merge-patch+json" }
        )
        return self._wrap_instance(self._extract_one(response))

    @flow
    def server_side_apply(
        self,
        id,
        data,
        /,
        field_manager = None,
        namespace = None,
        force = False
    ):
        """
        Uses server-side apply to create or update the specified object.

        See https://kubernetes.io/docs/reference/using-api/server-side-apply/.
        """
        field_manager = field_manager or self._client.default_field_manager
        namespace = namespace or data.get("metadata", {}).get("namespace")
        yield self._ensure_initialised()
        params = { "namespace": namespace, "fieldManager": field_manager }
        if force:
            params["force"] = "true"
        path, params = self._prepare_path(id, params)
        data = self._prepare_data(data, id, params)
        response = yield self._client.patch(
            path,
            json = data,
            params = params,
            # Use the special server-side apply content type
            headers = { "Content-Type": "application/apply-patch+yaml" }
        )
        return self._wrap_instance(self._extract_one(response))

    @flow
    def create_or_replace(self, id, data, /, namespace = None):
        # This is intended to replicate "kubectl apply"
        # So we fetch the latest resourceVersion before executing if required
        resource_version = data.get("metadata", {}).get("resourceVersion")
        if not resource_version:
            try:
                latest = yield self.fetch(id, namespace = namespace)
            except ApiError as exc:
                if exc.response.status_code == 404:
                    return (yield self.create(data, namespace = namespace))
                else:
                    raise
            else:
                latest_version = latest["metadata"]["resourceVersion"]
                data.setdefault("metadata", {})["resourceVersion"] = latest_version
        return (yield self.replace(id, data, namespace = namespace))

    def create_or_patch(self, id, data, /, namespace = None):
        return super().create_or_patch(id, data, namespace = namespace)

    def delete(
        self,
        id,
        /,
        propagation_policy = DeletePropagationPolicy.BACKGROUND,
        namespace = None
    ):
        if not isinstance(propagation_policy, DeletePropagationPolicy):
            propagation_policy = DeletePropagationPolicy(propagation_policy)
        return super().delete(
            id,
            {
                "apiVersion": "v1",
                "kind": "DeleteOptions",
                "propagationPolicy": propagation_policy.value,
            },
            namespace = namespace
        )

    @flow
    def delete_all(
        self,
        /,
        propagation_policy = DeletePropagationPolicy.BACKGROUND,
        **params
    ):
        """
        Deletes a collection of resources.
        """
        if not isinstance(propagation_policy, DeletePropagationPolicy):
            propagation_policy = DeletePropagationPolicy(propagation_policy)
        yield self._ensure_initialised()
        path, params = self._prepare_path(params = params)
        yield self._client.delete(
            path,
            json = {
                "apiVersion": "v1",
                "kind": "DeleteOptions",
                "propagationPolicy": propagation_policy.value,
            },
            params = params
        )

    @flow
    def watch_list(self, **params):
        """
        Watches a set of resource instances, as specified by the given parameters, for changes.

        Returns a tuple of (initial state, watch events).
        """
        yield self._ensure_initialised()
        # Accumulate the inital state by looping through the list iterator
        iterator = self.list(**params)
        initial_state = []
        while True:
            try:
                next_item = yield iterator._next_item()
            except ListResponseIterator.StopIteration:
                break
            else:
                initial_state.append(next_item)
        # Get the path to use for the watch
        path, params = self._prepare_path(params = params)
        # Use the final resource version from the iterator for the watch
        return initial_state, WatchEvents(self._client, path, params, iterator.resource_version)

    @flow
    def watch_one(self, id, /, namespace = None):
        """
        Watches a single resource instance for changes.

        Returns a tuple of (initial state, async iterator of watch events).
        """
        # Just watch the list but with a field selector
        # We also extract the single object from the initial state
        initial_state, events = yield self.watch_list(
            fields = { "metadata.name": id },
            namespace = namespace
        )
        return next(iter(initial_state), None), events



import contextlib
import json
import logging

from ..flow import Flowable, flow


logger = logging.getLogger(__name__)


class ListResponseIterator(Flowable):
    """
    Iterator for list responses.

    Can be used as either a sync or async iterator depending on the client that is given.
    """
    class StopIteration(Exception):
        """
        Exception raised to indicate that a list response has finished iterating.
        """

    def __init__(self, client, resource, params):
        self._client = client
        self._resource = resource
        self._data = []
        self._next_index = 0
        self._next_url, self._next_params = resource._prepare_path(params = params)

    def get_flow_executor(self):
        """
        Returns the flow executor to use.
        """
        return self._client.get_flow_executor()

    def _extract_list(self, response):
        """
        Extracts a list of instances from a list response.

        By default, this is deferred to the resource method so cases where a custom iterator
        is required are rare.
        """
        return self._resource._extract_list(response)

    def _extract_next_page(self, response):
        """
        Return a (URL, params) tuple for the next page or None if there is no next page.

        By default, this is deferred to the resource method so cases where a custom iterator
        is required are rare.
        """
        return self._resource._extract_next_page(response)

    @flow
    def _next_item(self):
        """
        Flow that returns the next item in the iterator.
        """
        yield self._resource._ensure_initialised()
        # If we have run out of data, try to load some more
        if self._next_index >= len(self._data) and self._next_url:
            response = yield self._client.get(self._next_url, params = self._next_params)
            self._data = self._extract_list(response)
            self._next_index = 0
            self._next_url, self._next_params = self._extract_next_page(response) or (None, None)
        # Return the item at the next index before incrementing it
        try:
            next_item = self._data[self._next_index]
        except IndexError:
            raise self.StopIteration
        else:
            self._next_index = self._next_index + 1
            return self._resource._wrap_instance(next_item)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self._next_item()
        except self.StopIteration:
            raise StopIteration

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return await self._next_item()
        except self.StopIteration:
            raise StopAsyncIteration


class StreamIterator:
    """
    Iterator over items in a stream.

    Can be used as a sync or async iterator depending on the client that is given.
    """
    class SuppressItem(Exception):
        """
        Exception that can be raised to suppress an item from being yielded.
        """

    def __init__(self, client, method, url, **kwargs):
        self._client = client
        self._method = method
        self._url = url
        # Split the given kwargs into request and send kwargs
        self._send_kwargs = {}
        if "auth" in kwargs:
            self._send_kwargs["auth"] = kwargs.pop("auth")
        if "follow_redirects" in kwargs:
            self._send_kwargs["follow_redirects"] = kwargs.pop("follow_redirects")
        self._request_kwargs = kwargs
        # Initially, there is no active response
        self._response = None

    def _request(self):
        """
        Returns the request for the iterator.
        """
        return self._client.build_request(self._method, self._url, **self._request_kwargs)

    @contextlib.contextmanager
    def _send(self):
        """
        Context manager that synchronously sends a request and yields a response, ensuring
        that the current response is updated.
        """
        self._response = self._client.send(
            self._request(),
            stream = True,
            **self._send_kwargs
        )
        try:
            yield self._response
        finally:
            self.close()

    @contextlib.asynccontextmanager
    async def _send_async(self):
        """
        Context manager that asynchronously sends a request and yields a response, ensuring
        that the current response is updated.
        """
        self._response = await self._client.send(
            self._request(),
            stream = True,
            **self._send_kwargs
        )
        try:
            yield self._response
        finally:
            await self.aclose()

    def _chunk_iterator(self, response):
        """
        Returns a synchronous chunk iterator for the response.
        """
        raise NotImplementedError

    def _async_chunk_iterator(self, response):
        """
        Returns an asynchronous chunk iterator for the response.
        """
        raise NotImplementedError

    def _process_chunk(self, chunk):
        """
        Process the given chunk of bytes and return the item to be yielded.

        To suppress the yielding of an item for a chunk, raise the special exception
        StreamIterator.SuppressItem.

        By default, chunks are processed as JSON.
        """
        return json.loads(chunk)

    def _should_resume(self, exception):
        """
        Receives the exception that caused the stream to exit, or None if the stream
        terminated without error, and returns a truthy value if the stream should be
        resumed.
        """
        # By default, this is a NOOP which causes the iterator to exit

    def __iter__(self):
        while True:
            try:
                with self._send() as response:
                    for chunk in self._chunk_iterator(response):
                        try:
                            yield self._process_chunk(chunk)
                        except self.SuppressItem:
                            continue
            except Exception as exc:
                if not self._should_resume(exc):
                    raise
            else:
                if not self._should_resume(None):
                    break

    async def __aiter__(self):
        while True:
            try:
                async with self._send_async() as response:
                    async for chunk in self._async_chunk_iterator(response):
                        try:
                            yield self._process_chunk(chunk)
                        except self.SuppressItem:
                            continue
            except Exception as exc:
                if not self._should_resume(exc):
                    raise
            else:
                if not self._should_resume(None):
                    break

    def close(self):
        if self._response:
            self._response.close()
            self._response = None

    async def aclose(self):
        if self._response:
            await self._response.aclose()
            self._response = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.aclose()


class ByteStreamIterator(StreamIterator):
    """
    Iterator for items in a byte stream.

    Chunks correspond to fixed-size chunks of a byte-stream, and are received by
    _process_chunk as bytes.
    """
    def __init__(self, client, method, url, /, chunk_size = None, **kwargs):
        super().__init__(client, method, url, **kwargs)
        self._chunk_size = chunk_size

    def _chunk_iterator(self, response):
        """
        Returns a synchronous chunk iterator for the response.
        """
        return response.iter_bytes(self._chunk_size)

    def _async_chunk_iterator(self, response):
        """
        Returns an asynchronous chunk iterator for the response.
        """
        return response.aiter_bytes(self._chunk_size)


class TextStreamIterator(StreamIterator):
    """
    Iterator for items in a text stream.

    Chunks correspond to lines of the response.
    """
    def _chunk_iterator(self, response):
        """
        Returns a synchronous chunk iterator for the response.
        """
        return response.iter_lines()

    def _async_chunk_iterator(self, response):
        """
        Returns an asynchronous chunk iterator for the response.
        """
        return response.aiter_lines()



import functools
import inspect


class SyncExecutor:
    """
    Flow executor for synchronous contexts.
    """
    def execute_flow(self, flow):
        """
        Executes a flow and returns the result.
        """
        try:
            action = flow.send
            to_send = None
            while True:
                try:
                    yielded_obj = action(to_send)
                except StopIteration as exc:
                    # If the generator returns, we are done
                    return exc.value
                try:
                    if inspect.isgenerator(yielded_obj):
                        to_send = self.execute_flow(yielded_obj)
                    else:
                        to_send = yielded_obj
                except Exception as exc:
                    action = flow.throw
                    to_send = exc
                else:
                    action = flow.send
        finally:
            flow.close()


class AsyncExecutor:
    """
    Flow executor for asynchronous contexts.
    """
    async def execute_flow(self, flow):
        """
        Executes a flow and returns an awaitable containing the result.
        """
        try:
            action = flow.send
            to_send = None
            while True:
                try:
                    yielded_obj = action(to_send)
                except StopIteration as exc:
                    # If the generator returns, we are done
                    return exc.value
                try:
                    if inspect.isgenerator(yielded_obj):
                        to_send = await self.execute_flow(yielded_obj)
                    elif inspect.isawaitable(yielded_obj):
                        to_send = await yielded_obj
                    else:
                        to_send = yielded_obj
                except Exception as exc:
                    action = flow.throw
                    to_send = exc
                else:
                    action = flow.send
        finally:
            flow.close()


class Flowable:
    """
    Base classes for objects that can support flows.
    """
    __flow_executor__ = None

    @property
    def is_async(self):
        return isinstance(self.__flow_executor__, AsyncExecutor)

    def get_flow_executor(self):
        return self.__flow_executor__


def flow(method):
    """
    Decorator that marks a method of a flowable as a flow that can be resolved synchronously
    or asynchronously depending on the executor that is in use.
    """
    @functools.wraps(method)
    def wrapper(flowable, *args, **kwargs):
        gen = method(flowable, *args, **kwargs)
        if inspect.isgenerator(gen):
            return flowable.get_flow_executor().execute_flow(gen)
        else:
            return gen
    return wrapper



import atexit
import base64
import os
import pathlib
import tempfile

import yaml

from .client import AsyncClient, SyncClient


class ConfigurationError(Exception):
    """
    Raised when there is an error with the configuration.
    """


def cleanup_tempfile(path):
    """
    Cleans up the specified file in a robust way.
    """
    try:
        os.remove(path)
    except OSError:
        pass


def file_or_data(obj, file_key, data_key = None):
    """
    Returns the path to a file containing the data for the specified key.
    """
    if file_key in obj:
        return obj[file_key]
    data_key = data_key or file_key + "-data"
    if data_key in obj:
        fd, path = tempfile.mkstemp()
        os.close(fd)
        # Register an exit handler to delete the file
        atexit.register(cleanup_tempfile, path)
        with open(path, 'wb') as fd:
            fd.write(base64.standard_b64decode(obj[data_key]))
        return path
    else:
        return None


class Configuration:
    """
    Configuration object that produces clients.
    """
    SA_HOST_ENV_NAME = "KUBERNETES_SERVICE_HOST"
    SA_PORT_ENV_NAME = "KUBERNETES_SERVICE_PORT"
    SA_TOKEN_FILENAME = "/var/run/secrets/kubernetes.io/serviceaccount/token"
    SA_CERT_FILENAME = "/var/run/secrets/kubernetes.io/serviceaccount/ca.crt"

    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def async_client(self, **kwargs):
        """
        Returns a new asynchronous client using the configuration.
        """
        merged = self._kwargs.copy()
        merged.update(kwargs)
        return AsyncClient(**merged)

    def sync_client(self, **kwargs):
        """
        Returns a new synchronous client using the configuration.
        """
        merged = self._kwargs.copy()
        merged.update(kwargs)
        return SyncClient(**merged)

    @classmethod
    def from_kubeconfig_data(cls, data, **kwargs):
        """
        Return a configuration for the given kubeconfig data, which can be bytes or str.
        """
        kubeconfig = yaml.safe_load(data)
        context = next(
            c["context"]
            for c in kubeconfig["contexts"]
            if c["name"] == kubeconfig["current-context"]
        )
        if "namespace" in context:
            kwargs.setdefault("default_namespace", context["namespace"])
        cluster = next(
            c["cluster"]
            for c in kubeconfig["clusters"]
            if c["name"] == context["cluster"]
        )
        kwargs.setdefault("base_url", cluster["server"])
        if "insecure-skip-tls-verify" in cluster:
            kwargs.setdefault("verify", False)
        else:
            ca_file = file_or_data(cluster, "certificate-authority")
            if ca_file:
                kwargs.setdefault("verify", ca_file)
        user = next(
            u["user"]
            for u in kubeconfig["users"]
            if u["name"] == context["user"]
        )
        client_cert = file_or_data(user, "client-certificate")
        if client_cert:
            kwargs.setdefault("cert", (client_cert, file_or_data(user, "client-key")))
        else:
            raise ConfigurationError("Authentication method not supported")
        return cls(**kwargs)

    @classmethod
    def from_kubeconfig(cls, path, **kwargs):
        """
        Return a configuration for the given kubeconfig file.
        """
        with pathlib.Path(path).open() as fh:
            data = fh.read()
        return cls.from_kubeconfig_data(data, **kwargs)

    @classmethod
    def from_serviceaccount(cls, **kwargs):
        """
        Return a configuration for the configured service account.
        """
        # The server and port come from environment variables
        server_host = os.environ.get(cls.SA_HOST_ENV_NAME)
        server_port = os.environ.get(cls.SA_PORT_ENV_NAME)
        if not server_host or not server_port:
            raise ConfigurationError("Server host/port not configured")
        kwargs.setdefault("base_url", f"https://{server_host}:{server_port}")
        # Check if the certificate file exists
        if os.path.isfile(cls.SA_CERT_FILENAME):
            kwargs.setdefault("verify", cls.SA_CERT_FILENAME)
        else:
            raise ConfigurationError("Service account CA file not present")
        try:
            with open(cls.SA_TOKEN_FILENAME) as fd:
                token = fd.read()
        except OSError:
            raise ConfigurationError("Could not read service account token")
        else:
            kwargs.setdefault("headers", {}).setdefault("authorization", f"bearer {token}")
        return cls(**kwargs)

    @classmethod
    def from_environment(cls, **kwargs):
        """
        Return a configuration from available information in the environment.
        """
        # An explicitly-specified kubeconfig file takes precedence
        if "KUBECONFIG" in os.environ:
            return cls.from_kubeconfig(os.environ["KUBECONFIG"], **kwargs)
        # Followed by a service account if configured
        try:
            return cls.from_serviceaccount(**kwargs)
        except ConfigurationError:
            pass
        # Lastly, the default kubeconfig location
        return cls.from_kubeconfig(pathlib.Path.home() / ".kube" / "config", **kwargs)



import dataclasses
import typing

from .resource import Resource


@dataclasses.dataclass
class ResourceSpec:
    """
    Specification for a Kubernetes API resource.
    """
    #: The API version, including group, of the resource
    api_version: str
    #: The name of the resource
    name: str
    #: The kind of the resource
    kind: typing.Optional[str] = None
    #: Whether or not the resource is namespaced
    namespaced: typing.Optional[bool] = None

    def __call__(self, client):
        """
        Return a resource instance for the given client.
        """
        return Resource(client, self.api_version, self.name, self.kind, self.namespaced)

    @classmethod
    def from_crd(cls, crd):
        """
        Returns a resource spec for the given CRD definition.
        """
        api_group = crd["spec"]["group"]
        preferred_version = next(
            version["name"]
            for version in crd["spec"]["versions"]
            if version.get("storage", False)
        )
        api_version = f"{api_group}/{preferred_version}"
        name = crd["spec"]["names"]["plural"]
        kind = crd["spec"]["names"]["kind"]
        namespaced = crd["spec"]["scope"] == "Namespaced"
        return cls(api_version, name, kind, namespaced)



import logging

import httpx

from ..flow import Flowable, flow

from .iterators import ListResponseIterator
from .util import PropertyDict


logger = logging.getLogger(__name__)


class Resource(Flowable):
    """
    Class for a REST resource.
    """
    __iterator_class__ = ListResponseIterator

    def __init__(self, client, name, prefix = None):
        self._client = client
        self._name = name.strip("/")
        self._prefix = prefix or "/"
        if not self._prefix.endswith("/"):
            self._prefix = self._prefix + "/"

    def get_flow_executor(self):
        """
        Returns the flow executor to use.
        """
        return self._client.get_flow_executor()

    def _ensure_initialised(self):
        """
        Ensures that the resource is fully initialised including any HTTP operations.

        This will be called multiple times, so should become a no-op in the case where
        the resource is already initialised.
        """
        # By default, this is a noop

    def _prepare_path(self, id = None, params = None):
        """
        Prepare the path to use for the given parameters.

        Returns a (path, params) tuple where the path parameters have interpolated
        and the remaining params should be given as URL parameters.
        """
        if id:
            parts = self._name.split("/", maxsplit = 1)
            parts.insert(1, id)
            return self._prefix + "/".join(parts), params
        else:
            return self._prefix + self._name, params

    def _prepare_data(self, data, id = None, params = None):
        """
        Prepare data for submitting as part of a create, replace or patch request.
        """
        return data

    def _extract_list(self, response):
        """
        Extracts a list of instances from a list response.
        """
        return response.json()

    def _extract_next_page(self, response):
        """
        Given a list response, return a (URL, params) tuple for the next page or None
        if there is no next page.
        """
        # By default, assume there is no pagination
        return None

    def _extract_one(self, response):
        """
        Extracts a single instance from a fetch, create or update response.
        """
        content_type = response.headers.get("content-type")
        if content_type == "application/json":
            return response.json()
        elif content_type == "text/plain":
            return response.text
        else:
            return response.content

    def _wrap_instance(self, instance):
        """
        Receives an instance and wraps the instance if required.
        """
        # By default, wrap the instance in a property dict
        return PropertyDict(instance) if isinstance(instance, dict) else instance

    def list(self, **params):
        """
        Returns an iterable of the resource instances that match the given parameters.
        """
        return self.__iterator_class__(self._client, self, params)

    @flow
    def first(self, **params):
        """
        Returns the first instance of a resource that matches the parameters, or None if one
        does not exist.
        """
        try:
            return (yield self.list(**params)._next_item())
        except ListResponseIterator.StopIteration:
            return None

    @flow
    def create(self, data, **params):
        """
        Creates an instance of the resource and returns it.
        """
        yield self._ensure_initialised()
        path, params = self._prepare_path(params = params)
        data = self._prepare_data(data, params = params)
        response = yield self._client.post(path, json = data, params = params)
        return self._wrap_instance(self._extract_one(response))

    @flow
    def fetch(self, id, **params):
        """
        Returns the data for the specified instance.
        """
        yield self._ensure_initialised()
        path, params = self._prepare_path(id, params)
        response = yield self._client.get(path, params = params)
        return self._wrap_instance(self._extract_one(response))

    @flow
    def replace(self, id, data, **params):
        """
        Replaces the specified instance with the given data.
        """
        yield self._ensure_initialised()
        path, params = self._prepare_path(id, params)
        data = self._prepare_data(data, id, params)
        response = yield self._client.put(path, json = data, params = params)
        return self._wrap_instance(self._extract_one(response))

    @flow
    def patch(self, id, data, **params):
        """
        Patches the specified instance with the given data.
        """
        yield self._ensure_initialised()
        path, params = self._prepare_path(id, params)
        data = self._prepare_data(data, id, params)
        response = yield self._client.patch(path, json = data, params = params)
        return self._wrap_instance(self._extract_one(response))

    def _create_or_update(self, method, id, data, params):
        """
        Flow that attempts to update an instance using the given data and method. If the
        instance does not exist, it is created with the given data.
        """
        # Always prepare the data with the id, even for a create
        data = self._prepare_data(data, id, params)
        try:
            return (yield method(id, data, **params))
        except httpx.HTTPStatusError as exc:
            if exc.response.status_code == 404:
                return (yield self.create(data, **params))
            else:
                raise

    @flow
    def create_or_replace(self, id, data, **params):
        """
        Attempts to replace the specified instance of the resource with the given data.
        If it does not exist, a new instance is created with the given data instead.
        """
        return (yield self._create_or_update(self.replace, id, data, params))

    @flow
    def create_or_patch(self, id, data, **params):
        """
        Attempts to patch the specified instance of the resource with the given data.
        If it does not exist, a new instance is created with the given data instead.
        """
        return (yield self._create_or_update(self.patch, id, data, params))

    @flow
    def delete(self, id, data = None, **params):
        """
        Delete the specified instance.
        """
        yield self._ensure_initialised()
        path, params = self._prepare_path(id, params)
        try:
            yield self._client.delete(path, json = data, params = params)
        except httpx.HTTPStatusError as exc:
            # Suppress 404s as the desired state has been reached
            if exc.response.status_code != 404:
                raise

    @flow
    def action(self, id, action, data = None, **params):
        """
        Executes an "action" for the specified instance.

        Executing an action means making a POST request to `/<resourcepath>/<id>/<action>`.
        """
        yield self._ensure_initialised()
        path, params = self._prepare_path(id, params)
        # Add the action onto the path, making care to preserve any trailing slashes
        action_path = f"{path}{action}/" if path.endswith("/") else f"{path}/{action}"
        response = yield self._client.post(action_path, json = data, params = params)
        content_type = response.headers.get("content-type")
        if content_type == "application/json":
            return response.json()
        elif content_type == "text/plain":
            return response.text
        else:
            return response.content



import asyncio
import json
import logging

import httpx

from ..flow import Flowable, flow, AsyncExecutor, SyncExecutor


logger = logging.getLogger(__name__)


class BaseClient(Flowable):
    """
    Base class for sync and async REST clients.
    """
    def __init__(self, /, json_encoder = None, **kwargs):
        super().__init__(**kwargs)
        self._json_encoder = json_encoder

    @flow
    def request(self, method, url, **kwargs):
        """
        Builds and sends a request, respecting any custom JSON encoder.
        """
        content = kwargs.get("content")
        json_obj = kwargs.pop("json", None)
        if content is None and json_obj is not None:
            kwargs["content"] = json.dumps(json_obj, default = self._json_encoder)
        return (yield super().request(method, url, **kwargs))

    @flow
    def send(self, request, **kwargs):
        """
        Sends the given request as part of a flow.
        """
        response = yield super().send(request, **kwargs)
        yield self.raise_for_status(response)
        return response

    @flow
    def raise_for_status(self, response):
        """
        Raise the relevant exception for the response, if required.
        """
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            # Make sure that the response is read while inside any required context managers
            if self.is_async:
                yield exc.response.aread()
            else:
                yield exc.response.read()
            raise exc

    @flow
    def delete(self, url, **kwargs):
        """
        Sends a delete request.

        We override the delete from HTTPX in order to be able to send a request body.
        """
        return (yield self.request("DELETE", url, **kwargs))


class SyncClient(BaseClient, httpx.Client):
    """
    Class for a sync REST client.
    """
    __flow_executor__ = SyncExecutor()


class AsyncClient(BaseClient, httpx.AsyncClient):
    """
    Class for a REST client.
    """
    __flow_executor__ = AsyncExecutor()

    async def _send_single_request(self, request):
        # anyio.fail_after seems to raise a TimeoutError even when the task is cancelled
        # from the outside rather than because it reaches the deadline
        # This is then translated into a PoolTimeout by httpx, even though the connection
        # pool has plenty of availability
        # We can prevent this while still allowing the request to appear cancelled from
        # the outside by shielding the request coroutine
        # This will mean that the request will continue until it is fulfilled, even if
        # nothing is waiting for it, but means PoolTimeouts will only be raised when
        # we genuinely fail to get a connection from the pool
        # As per the advice in the docs for shield, we maintain a reference to the
        # wrapped coroutine to avoid it being garbage collected too early
        coro = super()._send_single_request(request)
        response = await asyncio.shield(coro)
        return response



from collections.abc import MutableMapping


class PropertyDict(dict):
    """
    Dictionary that also supports property access.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(
            {
                k: self._wrap(v)
                for k, v in dict(*args, **kwargs).items()
            }
        )

    def _wrap(self, value):
        if isinstance(value, PropertyDict):
            return value
        elif isinstance(value, dict):
            return PropertyDict(value)
        else:
            return value

    def __setitem__(self, key, value):
        super().__setitem__(key, self._wrap(value))

    def setdefault(self, key, default = None):
        return super().setdefault(key, self._wrap(default))

    def update(self, *args, **kwargs):
        super().update(
            {
                k: self._wrap(v)
                for k, v in dict(*args, **kwargs).items()
            }
        )

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(f"'{self.__class__.__name__}' has no attribute '{name}'")

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()})"



import json

import httpx


class ApiError(httpx.HTTPStatusError):
    """
    Exception that is raised when a Kubernetes API error occurs that is in the 4xx range.
    """
    def __init__(self, source):
        try:
            data = source.response.json()
        except (json.JSONDecodeError, KeyError):
            message = f"{source.response.status_code} {source.response.reason_phrase}"
            reason = source.response.text
        else:
            message = data["message"]
            reason = data.get("reason")
        super().__init__(message, request = source.request, response = source.response)
        self.status_code = source.response.status_code
        self.reason = reason



from .api import Api
from .client import AsyncClient, SyncClient
from .errors import ApiError
from .iterators import ListResponseIterator, WatchEvents
from .resource import PRESENT, ABSENT, DeletePropagationPolicy, Resource
from .spec import ResourceSpec



import httpx

from ... import rest
from ...flow import flow

from .api import Api
from .errors import ApiError


class BaseClient:
    """
    Base class for sync and async clients.
    """
    def __init__(
        self,
        /,
        # This is the name of the field manager for server-side apply
        default_field_manager = "easykube",
        default_namespace = "default",
        **kwargs
    ):
        super().__init__(**kwargs)
        self.default_field_manager = default_field_manager
        self.default_namespace = default_namespace
        # Cache of API version -> API objects
        self.apis = {}
        # Cache of API group -> API version using preferred version
        self.preferred_versions = {}

    @flow
    def raise_for_status(self, response):
        # Convert response errors into ApiErrors for better messages
        try:
            yield super().raise_for_status(response)
        except httpx.HTTPStatusError as source:
            raise ApiError(source)

    def build_request(self, method, url, **kwargs):
        # For patch requests, set the content-type as merge-patch unless otherwise specified
        if method.lower() == "patch":
            headers = kwargs.get("headers") or {}
            headers.setdefault("Content-Type", "application/merge-patch+json")
            kwargs["headers"] = headers
        return super().build_request(method, url, **kwargs)

    def api(self, api_version):
        """
        Returns an API object for the specified API version.
        """
        if api_version not in self.apis:
            self.apis[api_version] = Api(self, api_version)
        return self.apis[api_version]

    @flow
    def api_preferred_version(self, group):
        """
        Returns an API object for the given group at the preferred version.
        """
        if group not in self.preferred_versions:
            response = yield self.get(f"/apis/{group}")
            api_version = response.json()["preferredVersion"]["groupVersion"]
            self.preferred_versions[group] = api_version
        return self.api(self.preferred_versions[group])

    def _resource_for_object(self, object):
        """
        Returns a resource for the given object.
        """
        # Get the resource using the api version and kind from the object
        api_version = object["apiVersion"]
        kind = object["kind"]
        return (yield self.api(api_version).resource(kind))

    @flow
    def create_object(self, object):
        """
        Create the given object.
        """
        resource = yield self._resource_for_object(object)
        namespace = object["metadata"].get("namespace")
        return (yield resource.create(object, namespace = namespace))

    @flow
    def replace_object(self, object):
        """
        Replace the given object with the specified state.
        """
        resource = yield self._resource_for_object(object)
        name = object["metadata"]["name"]
        namespace = object["metadata"].get("namespace")
        return (yield resource.replace(name, object, namespace = namespace))

    @flow
    def patch_object(self, object, patch):
        """
        Apply the given patch to the specified object.
        """
        resource = yield self._resource_for_object(object)
        name = object["metadata"]["name"]
        namespace = object["metadata"].get("namespace")
        return (yield resource.patch(name, patch, namespace = namespace))

    @flow
    def delete_object(self, object):
        """
        Delete the specified object.
        """
        resource = yield self._resource_for_object(object)
        name = object["metadata"]["name"]
        namespace = object["metadata"].get("namespace")
        return (yield resource.delete(name, namespace = namespace))

    @flow
    def apply_object(self, object, /, field_manager = None, force = False):
        """
        Applies the given object using server-side apply.

        See https://kubernetes.io/docs/reference/using-api/server-side-apply/.
        """
        resource = yield self._resource_for_object(object)
        name = object["metadata"]["name"]
        namespace = object["metadata"].get("namespace")
        return (
            yield resource.server_side_apply(
                name,
                object,
                field_manager = field_manager,
                namespace = namespace,
                force = force
            )
        )

    @flow
    def client_side_apply_object(self, object):
        """
        Create or update the given object, equivalent to "kubectl apply".
        """
        resource = yield self._resource_for_object(object)
        name = object["metadata"]["name"]
        namespace = object["metadata"].get("namespace")
        return (yield resource.create_or_replace(name, object, namespace = namespace))


class SyncClient(BaseClient, rest.SyncClient):
    """
    Sync client for Kubernetes.
    """


class AsyncClient(BaseClient, rest.AsyncClient):
    """
    Async client for Kubernetes.
    """
